""" Ableton `.alc` files to json """
import os
import json
import gzip
from bs4 import BeautifulSoup


def make_json_dir(directory, out_path):
    """
        Creates a directory of json files extracted from `.alc` files.
    :param directory: Path of `.alc` files to be converted.
    :param out_path: Json output path.
    :return:
    """
    for alc_file in directory:
        try:
            extract_alc_features(alc_file, out_path)
        except Exception as e:
            print(f'Failed for {alc_file}\nWith Exception : {e}')


def alc_folder_path(folder, kit=True):
    """
        Returns a list of all `.alc` filepaths from a given directory.
    :param kit: If True will only return paths that contain `KIT` in filename. (Drumkits)
    :param folder: Path of `.alc` files folder
    :return: List of full `.alc` filepaths
    """

    dirs = [x[0] for x in os.walk(folder)]
    dirs = [x for x in dirs if 'MIDI Clips\\' in x]
    alc_paths = []
    for _dir in dirs:
        for alc in os.listdir(_dir):
            if '.alc' in alc and '.ogg' not in alc:
                alc_paths.append(_dir + '\\' + alc)
    if kit:
        alc_paths = [alc for alc in alc_paths if 'KIT' in alc.upper().split('\\')[-1]]

    return alc_paths


def extract_alc_features(alc_file, out_path, drumkit=True):
    """
        Extracts midi features from a `.alc` file and creates a json output file.
    :param alc_file: Path of Ableton alc file. Example -> C:/Users/USERNAME/Documents/Ableton/Factory Packs'
    :param out_path: Path of exported json file.
    :param drumkit: If False full range of midi notes will be processed.
    :return:
    """

    if drumkit:
        key_notes_midi = [str(x) for x in range(36, 52)]
    else:
        key_notes_midi = [str(x) for x in range(128)]
    # Filename.
    fname = alc_file.split('\\')[-1].split('.')[0]
    # Open file with gzip, parse with bs4.
    with gzip.open(alc_file, 'rb') as f:
        file_content = f.read()
        file_content = file_content.decode("utf-8")
        file_content = BeautifulSoup(file_content, 'lxml')
    # Get the midi notes.
    notes = file_content.ableton.liveset.tracks.miditrack.devicechain.mainsequencer.clipslotlist.clipslot.clipslot.value.midiclip.notes
    # Create data pairs for note, time, duration, & velocity.
    preset_pairs = []
    for midi in notes.findAll('keytrack'):
        if drumkit:
            if 36 <= int(midi.midikey['value']) < 52:
                _midi = midi.findAll('midinoteevent')
                midi_group = []
                for _notes in _midi:
                    midi_group.append([_notes['time'], _notes['duration'], _notes['velocity']])
                midi_key = midi.midikey['value']
                preset_pairs.append([midi_key, midi_group])
                key_notes_midi.remove(midi_key)
        else:
            _midi = midi.findAll('midinoteevent')
            midi_group = []
            for _notes in _midi:
                midi_group.append([_notes['time'], _notes['duration'], _notes['velocity']])
            midi_key = midi.midikey['value']
            preset_pairs.append([midi_key, midi_group])
            key_notes_midi.remove(midi_key)
    # Create preset pairs.
    for _m in key_notes_midi:
        preset_pairs.append([_m, []])
    # Create a dictionary for the preset.
    _dict = {}
    for items in preset_pairs:
        _dict[items[0]] = items[1]
    _dict = dict(sorted(_dict.items()))
    # Export json file.
    with open(f'{out_path}{os.sep}{fname}', 'w', encoding='utf-8') as f:
        json.dump(_dict, f, ensure_ascii=False, indent=4)



""" Create midi file from a combined pattern of `.alc` files """
import os
import json
import random
import note_seq
from note_seq.protobuf import music_pb2
# Randomness
SYS_RAND = random.SystemRandom()


def generate_pattern(json_dirs, out_path=None, fname='python_als_gen_def_test', bars=1, fills=False):
    """
        Generates a midi pattern from 16 random kits.
        Additional option to add fills at the end of every 2 bars.
    :param json_dirs: Path of generated json files.
    :param out_path: Path for output midi files.
    :param fname: Midi filename.
    :param bars: Number of bars in sequence.
    :param fills: If True generate drum fills.
    :return: Exports `.mid` file.
    """
    # List of `.alc` files.
    jsons = os.listdir(json_dirs)
    # Chose 16 random kits.
    random_kits = [SYS_RAND.choice(jsons) for x in range(16)]
    # Blank midi file.
    midi_outp = music_pb2.NoteSequence()
    # Fill to be added every 2 bars if bars not 1.
    fill_kits = int(bars / 2)
    # If fills selected create fill midi sequence from kits.
    if fills and bars > 1:
        bar = 1
        for fill in range(fill_kits):
            fkits = [SYS_RAND.choice(jsons) for x in range(16)]
            for n, rando in enumerate(fkits, start=36):
                with open(f'{json_dirs}{os.sep}{rando}', 'r') as f:
                    array = json.load(f)
                _notes = array['{}'.format(n)]
                if _notes:
                    for _m_note in _notes:
                        if 6 < float(_m_note[0]) < 7.95:
                            midi_outp.notes.add(pitch=n,
                                                start_time=(float(_m_note[0]) / 2) + (bar * 4),
                                                end_time=((float(_m_note[0]) + float(_m_note[1])) / 2) + (bar * 4),
                                                velocity=int(float(_m_note[2])))
            bar += 2
    # Create midi sequence from kits.
    for n, rando in enumerate(random_kits, start=36):
        with open(f'{json_dirs}{os.sep}{rando}', 'r') as f:
            array = json.load(f)
        _notes = array['{}'.format(n)]
        if _notes:
            for _m_note in _notes:
                bar = 0
                for _bar in range(bars):
                    for _m_note in _notes:
                        if fills and bar % 2 != 0:
                            if float(_m_note[0]) < 5.95:
                                midi_outp.notes.add(pitch=n,
                                                    start_time=(float(_m_note[0]) / 2) + (bar * 4),
                                                    end_time=((float(_m_note[0]) + float(_m_note[1])) / 2) + (bar * 4),
                                                    velocity=int(float(_m_note[2])))
                        else:
                            if float(_m_note[0]) < 7.95:
                                midi_outp.notes.add(pitch=n,
                                                    start_time=(float(_m_note[0]) / 2) + (bar * 4),
                                                    end_time=((float(_m_note[0]) + float(_m_note[1])) / 2) + (bar * 4),
                                                    velocity=int(float(_m_note[2])))
                    bar += 1
    # Export path.
    exp_path = out_path + os.sep + fname + '.mid' if out_path else fname + '.mid'
    # Export midi file.
    note_seq.sequence_proto_to_midi_file(midi_outp, exp_path)



from setuptools import setup

setup(
	name='alcmixer',
	version='0.1.0',
	author='Daniel Kavoulakos',
	author_email='dan_kavoulakos@hotmail.com',
	description='Create midi drum patterns by randomly combining multiple Ableton Live Clips',
	license='MIT',
	packages=['.alcmixer'],
	url = 'https://github.com/danikavu/Ableton-Live-Clip-Mixer',
	install_requires=[
		'note_seq',
		'beautifulsoup4',
		],
	keywords=['Ableton', 'drumkit', 'midi', 'alc'],
	python_requires='>=3.6',

)


# -*- pretty cool init mate? -*-
from alcmixer.generate_midi import generate_pattern
from alcmixer.extract_alc_features import extract_alc_features, alc_folder_path, make_json_dir

__all__ = [
    'generate_pattern',
    'extract_alc_features',
    'alc_folder_path',
    'make_json_dir',
    ]